from ._ks import CalculatedKS
from ._lift import CalculatedLift
from ._roc_auc import CalculatedAUC


def ks(y_pred, y_target):
    return CalculatedKS(y_pred, y_target).ks()


def plot_ks_chart(y_pred, y_target):
    return CalculatedKS(y_pred, y_target).plot_ks_chart()


def lift(y_pred, y_target, bucket=10):
    return CalculatedLift(y_pred, y_target, bucket).lift()


def plot_gain_chart(y_pred, y_target, bucket=10):
    return CalculatedLift(y_pred, y_target, bucket).plot_gain_chart()


def plot_lift_chart(y_pred, y_target, bucket=10):
    return CalculatedLift(y_pred, y_target, bucket).plot_lift_chart()


def gini(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).gini()


def roc_auc(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).roc_auc()


def plot_roc_curve(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).plot_roc_curve()


def plot_compare_true_label(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).plot_compare_true_label()


def plot_tpr_fpr(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).plot_tpr_fpr()


def plot_pr_curve(y_pred, y_target):
    return CalculatedAUC(y_pred, y_target).plot_pr_curve()



import numpy as np
import pandas as pd
import plotly.express as px

from credit_scoring.metrics.credit_score import CreditScore


class CalculatedKS(CreditScore):

	def ks(self) -> tuple:
		'''
		Calculate the KS table of each decile, KS value, the deicile at KS
		:return: KS Tabel, KS value, Decile
		'''

		# Create the dataframe contain target labels and probability values
		data = pd.DataFrame({"target": self.target, "prob": self.pred})

		# Calculate the probability of nonvents
		data['target0'] = 1 - data['target']

		# Find the suitable bucket
		self.bucket = min(self.bucket, data['prob'].nunique())

		# Cut the probability values to 'bucket' ranges by quantile
		data['bucket'] = pd.qcut(data['prob'], self.bucket)
		grouped = data.groupby('bucket', as_index=False)

		ks_table = pd.DataFrame()

		# Calculate the min, max of probability,
		ks_table['min_prob'] = grouped.min()['prob']
		ks_table['max_prob'] = grouped.max()['prob']

		# Calculate the number of events, nonevents
		ks_table['events'] = grouped.sum()['target']
		ks_table['nonevents'] = grouped.sum()['target0']

		# Calculate the cumulative probability of event and non-events in each decile
		ks_table = ks_table.sort_values(by="min_prob", ascending=False).reset_index(drop=True)
		ks_table['cum_eventrate'] = (ks_table.events / data['target'].sum()).cumsum()
		ks_table['cum_noneventrate'] = (ks_table.nonevents / data['target0'].sum()).cumsum()

		# CAlculate the KS by formula
		ks_table['KS'] = np.round(ks_table['cum_eventrate'] - ks_table['cum_noneventrate'], 3) * 100
		ks_table.loc[0, 'cum_eventrate'] = ks_table.loc[0, 'cum_noneventrate'] = 0
		ks_table.index = range(self.bucket)
		ks_table.index.rename('Decile', inplace=True)

		ks_val, decile = max(ks_table['KS']), ks_table.index[ks_table['KS'] == max(ks_table['KS'])][0]
		return ks_table, ks_val, decile

	def plot_ks_chart(self) -> None:
		"""
		Plot the KS chart, distribution of cdf of event and non-event
		:return: None
		"""
		data, ks_val, decile = self.ks()
		text = str(ks_val) + "%" + " at decile " + str(decile)
		fig = px.line(data, x=data.index, y=['cum_eventrate', 'cum_noneventrate'],
		              title='Kolmogorov-Smirnov statistic: ' + text)
		fig.add_vline(x=decile, line_width=3, line_dash="dash", line_color="green")
		fig.show()


from .metrics import (
    ks,
    plot_ks_chart,
    lift,
    plot_gain_chart,
    plot_lift_chart,
    gini,
    roc_auc,
    plot_roc_curve,
    plot_compare_true_label,
    plot_tpr_fpr,
    plot_pr_curve
)



import pandas as pd
import plotly.express as px
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve, auc, precision_recall_curve

from credit_scoring.metrics.credit_score import CreditScore


class CalculatedAUC(CreditScore):

	def gini(self) -> float:
		"""
		Calculated the gini value base on prediction values and target values
		:return: GINI value
		"""
		return 2 * roc_auc_score(self.target, self.pred) - 1

	def roc_auc(self) -> float:
		"""
		Calculated the roc_auc value base on prediction values and target values
		:return: The ROC_AUC value
		"""
		return roc_auc_score(self.target, self.pred)

	def plot_roc_curve(self) -> None:
		"""
		Using plotly to plot roc curve
		:return: None
		"""
		fpr, tpr, thresholds = roc_curve(self.target, self.pred)
		fig = px.area(
			x=fpr, y=tpr,
			title=f'ROC Curve (AUC={auc(fpr, tpr):.4f})',
			labels=dict(x='False Positive Rate', y='True Positive Rate'),
			width=700, height=500
		)
		fig.add_shape(
			type='line', line=dict(dash='dash'),
			x0=0, x1=1, y0=0, y1=1
		)

		fig.update_yaxes(scaleanchor="x", scaleratio=1)
		fig.update_xaxes(constrain='domain')
		fig.show()

	def plot_compare_true_label(self) -> None:
		"""
		Using plotly to display how the true labels better than false labels
		:return: None
		"""
		fig_hist = px.histogram(
			x=self.pred, color=self.target, nbins=50,
			labels=dict(color='True Labels', x='Score')
		)
		fig_hist.show()

	def plot_tpr_fpr(self) -> None:
		"""
		Using plotly to plot the relation between TPR and FPR at every threshold
		:return: None
		"""
		fpr, tpr, thresholds = roc_curve(self.target, self.pred)
		df = pd.DataFrame({
			'False Positive Rate': fpr,
			'True Positive Rate': tpr
		}, index=thresholds)
		df.index.name = "Thresholds"
		df.columns.name = "Rate"

		fig_thresh = px.line(
			df, title='TPR and FPR at every threshold',
			width=700, height=500
		)
		fig_thresh.update_yaxes(scaleanchor="x", scaleratio=1)
		fig_thresh.update_xaxes(range=[0, 1], constrain='domain')
		fig_thresh.show()

	def plot_pr_curve(self) -> None:
		"""
		Using plotly to display the PR curve
		:return: None
		"""
		precision, recall, _ = precision_recall_curve(self.target, self.pred)
		fpr, tpr, _ = roc_curve(self.target, self.pred)
		fig = px.area(
			x=recall, y=precision,
			title=f'Precision-Recall Curve (AUC={auc(fpr, tpr):.4f})',
			labels=dict(x='Recall', y='Precision'),
			width=700, height=500
		)
		fig.add_shape(
			type='line', line=dict(dash='dash'),
			x0=0, x1=1, y0=1, y1=0
		)
		fig.update_yaxes(scaleanchor="x", scaleratio=1)
		fig.update_xaxes(constrain='domain')

		fig.show()



import numpy as np
import pandas as pd
import plotly.graph_objects as go

from credit_scoring.metrics.credit_score import CreditScore


class CalculatedLift(CreditScore):
	def __init__(self, pred, target, bucket=10):
		super().__init__(pred, target)
		self.bucket = bucket
		self.pred = 1 - self.pred

	def lift(self) -> pd.DataFrame:
		"""

		:return: The lift table contains Gain, Lift each deciles
		"""

		# Create dataframe contain target and prediction values
		data = pd.DataFrame({'Target': self.target, 'Pred': self.pred})
		data = data.sort_values(by='Pred', ascending=False)

		# Cut Prediction values to q range by percentile
		data['decile'] = pd.qcut(data['Pred'], q=self.bucket)

		# Calculate number events and total records, PS: the bellow code only support pandas from version 1.0.0
		lift = data.groupby('decile').agg(
			Number_of_customers=('Target', 'count'),
			Number_of_goods=('Target', 'sum')
		)

		# Calculate cumulative good by sumup in each decile
		lift['Cumulative_goods'] = lift['Number_of_goods'].cumsum()

		# Calculate the distribution of Events in each decile
		lift['Percent_of_Events'] = lift['Number_of_goods'] / lift['Number_of_goods'].sum() * 100

		# Calculate Gain value by sumup of events distribution
		lift['Gain'] = lift['Percent_of_Events'].cumsum()

		# Calculate the random probability of events in each decile
		sample = [i * 100 / self.bucket for i in range(1, self.bucket + 1)]

		# The lift would be : Gain / random probability of each decile
		lift['Lift'] = lift['Gain'] / np.array(sample)
		return lift

	def plot_gain_chart(self) -> None:
		"""

		:return: Plot the Gain chart base on the result of lift function above
		"""
		data = self.lift()
		gain = data.Gain.tolist()
		gain.insert(0, 0)
		fig = go.Figure()
		fig.add_trace(go.Scatter(x=list(np.linspace(0, 100, self.bucket + 1, dtype=int)),
		                         y=list(np.linspace(0, 100, self.bucket + 1, dtype=int)),
		                         mode='lines+markers',
		                         name='lines+markers'))
		fig.add_trace(go.Scatter(x=list(np.linspace(0, 100, self.bucket + 1, dtype=int)), y=gain,
		                         mode='lines+markers',
		                         name='lines+markers'))
		fig.update_xaxes(title_text="Percentage")
		fig.update_yaxes(title_text="% of Gain")
		fig.update_layout(title='Gain Charts')
		fig.show()

	def plot_lift_chart(self) -> None:
		"""

		:return: Plot the Lift chart base on the reuslt of lift function above
		"""
		data = self.lift()
		fig = go.Figure()
		fig.add_trace(go.Scatter(x=list(np.linspace(0, 100, self.bucket + 1, dtype=int)),
		                         y=np.repeat(1, self.bucket),
		                         mode='lines+markers',
		                         name='lines+markers'))
		fig.add_trace(go.Scatter(x=list(np.linspace(0, 100, self.bucket + 1, dtype=int)),
		                         y=data.Lift,
		                         mode='lines+markers',
		                         name='lines+markers'))
		fig.update_xaxes(title_text="Percentage")
		fig.update_yaxes(title_text="Lift")
		fig.update_layout(title='Lift Charts')
		fig.show()



import pandas as pd
import numpy as np
from collections import defaultdict


class WOE:
    """
    Process the WOE-IV for dataframe
    """

    def __init__(self, frame: pd.DataFrame, target: str,
                 min_unique_to_combine: int, is_combine=False):
        """

        :param frame: the dataframe contains features (only features need to be woe) and target (output)
        :param target: the label
        """
        self.frame = frame
        self.target = target
        self.is_combine = is_combine
        self.min_unique_to_combine = min_unique_to_combine
        self.check_type()

    def check_type(self):
        """
        Check valid input
        :return:
        """
        # the first input must be dataframe
        if not isinstance(self.frame, pd.DataFrame):
            msg_error = "The first input must be dataframe"
            raise ValueError(msg_error)

        # check the target label is in input dataframe
        if self.target not in self.frame.columns:
            msg_error = "The target label '{}' must be in input data-frame".format(self.target)
            raise ValueError(msg_error)

        # check the target label is binary type
        if self.frame[self.target].nunique() != 2:
            msg_error = "The label target is not binary type."
            raise ValueError(msg_error)

        # check the target only contains 0 and 1,
        if sorted(list(self.frame[self.target].unique())) != [0, 1]:
            msg_error = "The label target should be encode to 0 and 1"
            raise ValueError(msg_error)

    def mono_woe(self, column):
        """
        Calculate woe for 1 features
        :return: woe table value
        """

        res = list()
        values = self.frame[column].unique()
        for value in values:
            tmp = self.frame[self.frame[column] == value]
            res.append(
                {
                    'Value': value,
                    'Occur': tmp.shape[0],
                    'Event': tmp.loc[tmp[self.target] == 1].shape[0],
                    'NonEvent': tmp.loc[tmp[self.target] == 0].shape[0],
                }
            )
        data = pd.DataFrame(res)
        data['Good_Distribution'] = data['NonEvent'] / data['NonEvent'].sum()
        data['Bad_Distribution'] = data['Event'] / data['Event'].sum()

        # WOE = Ln(% non-event / % event)
        data['WOE'] = np.log(data['Good_Distribution'] / data['Bad_Distribution'])

        # Replace infinity to 0, sometime the bad distribution equal 0
        data = data.replace({'WOE': {np.inf: 0, -np.inf: 0}})

        # IV = (% non-event - %event) * woe
        data['IV'] = (data['Good_Distribution'] - data['Bad_Distribution']) * data['WOE']
        iv = data['IV'].sum()
        data = data.sort_values(by=['WOE'])

        # check combine flag
        group = None
        if self.is_combine and self.min_unique_to_combine < len(values):
            data, group = self.auto_combined(data, self.min_unique_to_combine)

        data['Feature'] = column
        return data, iv, group

    @staticmethod
    def auto_combined(frame, q):
        min_, max_ = frame['WOE'].min() - 1e-3, frame['WOE'].max() + 1e-3
        range_ = np.linspace(min_, max_, q + 1)
        frame['TMP'] = pd.cut(frame['WOE'], range_)
        values = frame['TMP'].unique()
        res, group = list(), defaultdict(set)

        for idx, value in enumerate(values):
            k = frame.loc[frame['TMP'] == value]
            k1 = pd.DataFrame(np.mean(k)).T
            name_group = 'GROUP_{}'.format(idx)
            k1['Value'] = name_group
            res.append(k1)
            group[name_group] |= set(k['Value'].unique())

        res = pd.concat(res)
        res = res.sort_values(by=['WOE'], ascending=False).reset_index(drop=True)
        return res, group

    def woe_iv(self):
        columns = set(self.frame.columns) ^ {self.target}
        woe, iv = list(), dict()

        # Replace NA value with String to handle na case
        self.frame.fillna("NA Value", inplace=True)

        res = {}
        for column in columns:
            t, r, _ = self.mono_woe(column)
            woe.append(t)
            iv[column] = r
            res[column] = _

        iv = pd.DataFrame([iv]).T
        woe = pd.concat(woe)
        return woe, iv, res



from distutils.core import setup
from os.path import exists
from setuptools import setup, find_packages

setup(
    name='credit_scoring',
    version='0.1.0',
    license='MIT',
    description='This library help researcher have quick result with woe, iv, '
                'result reporting with credit scoring problem',
    # long_description=open('README.md').read() if exists("README.md") else "",
    author='Minh Tran',
    author_email='minhtc.uet@gmail.com',
    url='https://github.com/minhtcuet/creditscoring',
    download_url='https://github.com/minhtcuet/creditscoring/archive/v_01.tar.gz',
    keywords=['Credit', 'Credit Score', 'WOE', 'IV', 'WOE-IV', 'GINI', 'KS', 'Lift', 'GAIN'],
    install_requires=[
        'numpy',
        'pandas',
        'plotly',
        'sklearn',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(exclude=['build', 'docs', 'templates']),

)


import numpy as np
import pandas as pd


class CreditScore:
	"""
	This class will calculated the KS value. (Only support binary classification - Credit Scoring for example)
	KS is defined as maximal absolute difference between CDFs of scores of good and bad clients

	PS: This version only support for continuous variable
	Random Variable will be able later version.
	"""

	def __init__(self, pred, target, bucket=200):
		"""

		:param pred: The value of prediction
		:param target: The target relevant
		"""
		self.pred = pred
		self.target = target
		self.check_valid_type()
		self.bucket = min(bucket, len(self.pred), len(self.target))

	def check_valid_type(self) -> bool:
		"""
		Check valid input type. Only supported np.array, pd.Seris, list
		:return:
		"""
		# Check supported type input
		if type(self.pred) not in [list, np.ndarray, pd.core.series.Series]:
			error_msg = "Your input type must be one of following: <class 'pandas.core.series.Series'>, " \
			            "<class 'numpy.ndarray'>, <class 'list'>"
			raise ValueError(error_msg)

		# Compare length
		if len(self.pred) != len(self.target):
			error_msg = "All arguments should have the same length."
			raise ValueError(error_msg)

		# Convert to numpy array and check the shape
		try:
			self.pred = np.array(self.pred, dtype=float)
			self.target = np.array(self.target, dtype=float)
			if len(self.pred.shape) != 1 or len(self.target.shape) != 1:
				error_msg = "Your inputs shape must be one demension"
				raise ValueError(error_msg)
		except Exception as e:
			raise ValueError(e)
		return True
